@@ -11,8 +11,8 @@
  * GNU General Public License for more details.
  */
 
-#ifndef	VERSION
-#define VERSION	"tig-0.4.git"
+#ifndef VERSION
+#define VERSION "unknown-version"
 #endif
 
 #ifndef DEBUG
@@ -60,6 +60,15 @@ static size_t utf8_length(const char *string, size_t max_width, int *coloffset,
 #define SIZEOF_STR	1024	/* Default string size. */
 #define SIZEOF_REF	256	/* Size of symbolic or SHA1 ID. */
 #define SIZEOF_REV	41	/* Holds a SHA-1 and an ending NUL */
+
+/* Revision graph */
+
+#define REVGRAPH_INIT	'I'
+#define REVGRAPH_MERGE	'M'
+#define REVGRAPH_BRANCH	'+'
+#define REVGRAPH_COMMIT	'*'
+#define REVGRAPH_LINE	'|'
+
 #define SIZEOF_REVGRAPH	19	/* Size of revision ancestry graphics. */
 
 /* This color name can be used to refer to the default term colors. */
@@ -81,7 +90,7 @@ static size_t utf8_length(const char *string, size_t max_width, int *coloffset,
 #define	SCALE_SPLIT_VIEW(height)	((height) * 2 / 3)
 
 #define TIG_LS_REMOTE \
-	"git ls-remote . 2>/dev/null"
+	"git ls-remote $(git rev-parse --git-dir) 2>/dev/null"
 
 #define TIG_DIFF_CMD \
 	"git show --root --patch-with-stat --find-copies-harder -B -C %s 2>/dev/null"
@@ -101,6 +110,7 @@ static size_t utf8_length(const char *string, size_t max_width, int *coloffset,
 /* XXX: Needs to be defined to the empty string. */
 #define TIG_HELP_CMD	""
 #define TIG_PAGER_CMD	""
+#define TIG_STATUS_CMD	""
 
 /* Some ascii-shorthands fitted into the ncurses namespace. */
 #define KEY_TAB		'\t'
@@ -112,6 +122,7 @@ struct ref {
 	char *name;		/* Ref name; tag or head names are shortened. */
 	char id[SIZEOF_REV];	/* Commit SHA1 ID */
 	unsigned int tag:1;	/* Is it a tag? */
+	unsigned int remote:1;	/* Is it a remote ref? */
 	unsigned int next:1;	/* For ref lists: are there more refs? */
 };
 
@@ -163,6 +174,9 @@ string_ncopy_do(char *dst, size_t dstlen, const char *src, size_t srclen)
 #define string_ncopy(dst, src, srclen) \
 	string_ncopy_do(dst, sizeof(dst), src, srclen)
 
+#define string_copy_rev(dst, src) \
+	string_ncopy_do(dst, SIZEOF_REV, src, SIZEOF_REV - 1)
+
 static char *
 chomp_string(char *name)
 {
@@ -259,6 +273,9 @@ sq_quote(char buf[SIZEOF_STR], size_t bufsize, const char *src)
 	}
 	BUFPUT('\'');
 
+	if (bufsize < SIZEOF_STR)
+		buf[bufsize] = 0;
+
 	return bufsize;
 }
 
@@ -277,6 +294,7 @@ sq_quote(char buf[SIZEOF_STR], size_t bufsize, const char *src)
 	REQ_(VIEW_BLOB,		"Show blob view"), \
 	REQ_(VIEW_HELP,		"Show help page"), \
 	REQ_(VIEW_PAGER,	"Show pager view"), \
+	REQ_(VIEW_STATUS,	"Show status view"), \
 	\
 	REQ_GROUP("View manipulation") \
 	REQ_(ENTER,		"Enter current line and scroll"), \
@@ -366,7 +384,7 @@ get_request(const char *name)
  */
 
 static const char usage[] =
-VERSION " (" __DATE__ ")\n"
+"tig " VERSION " (" __DATE__ ")\n"
 "\n"
 "Usage: tig [options]\n"
 "   or: tig [options] [--] [git log options]\n"
@@ -378,6 +396,7 @@ VERSION " (" __DATE__ ")\n"
 "Options:\n"
 "  -l                          Start up in log view\n"
 "  -d                          Start up in diff view\n"
+"  -S                          Start up in status view\n"
 "  -n[I], --line-number[=I]    Show line numbers with given interval\n"
 "  -b[N], --tab-size[=N]       Set number of spaces for tab expansion\n"
 "  --                          Mark end of tig options\n"
@@ -386,7 +405,7 @@ VERSION " (" __DATE__ ")\n"
 
 /* Option and state variables. */
 static bool opt_line_number		= FALSE;
-static bool opt_rev_graph		= TRUE;
+static bool opt_rev_graph		= FALSE;
 static int opt_num_interval		= NUMBER_INTERVAL;
 static int opt_tab_size			= TABSIZE;
 static enum request opt_request		= REQ_VIEW_MAIN;
@@ -472,6 +491,11 @@ parse_options(int argc, char *argv[])
 			continue;
 		}
 
+		if (!strcmp(opt, "-S")) {
+			opt_request = REQ_VIEW_STATUS;
+			break;
+		}
+
 		if (check_option(opt, 'n', "line-number", OPT_INT, &opt_num_interval)) {
 			opt_line_number = TRUE;
 			continue;
@@ -510,7 +534,7 @@ parse_options(int argc, char *argv[])
 		if (opt_request == REQ_VIEW_MAIN)
 			/* XXX: This is vulnerable to the user overriding
 			 * options required for the main view parser. */
-			string_copy(opt_cmd, "git log --stat --pretty=raw");
+			string_copy(opt_cmd, "git log --pretty=raw");
 		else
 			string_copy(opt_cmd, "git");
 		buf_size = strlen(opt_cmd);
@@ -524,7 +548,6 @@ parse_options(int argc, char *argv[])
 			die("command too long");
 
 		opt_cmd[buf_size] = 0;
-
 	}
 
 	if (*opt_encoding && strcasecmp(opt_encoding, "UTF-8"))
@@ -577,9 +600,15 @@ LINE(MAIN_AUTHOR,  "",			COLOR_GREEN,	COLOR_DEFAULT,	0), \
 LINE(MAIN_COMMIT,  "",			COLOR_DEFAULT,	COLOR_DEFAULT,	0), \
 LINE(MAIN_DELIM,   "",			COLOR_MAGENTA,	COLOR_DEFAULT,	0), \
 LINE(MAIN_TAG,     "",			COLOR_MAGENTA,	COLOR_DEFAULT,	A_BOLD), \
+LINE(MAIN_REMOTE,  "",			COLOR_YELLOW,	COLOR_DEFAULT,	A_BOLD), \
 LINE(MAIN_REF,     "",			COLOR_CYAN,	COLOR_DEFAULT,	A_BOLD), \
 LINE(TREE_DIR,     "",			COLOR_DEFAULT,	COLOR_DEFAULT,	A_NORMAL), \
-LINE(TREE_FILE,    "",			COLOR_DEFAULT,	COLOR_DEFAULT,	A_NORMAL)
+LINE(TREE_FILE,    "",			COLOR_DEFAULT,	COLOR_DEFAULT,	A_NORMAL), \
+LINE(STAT_SECTION, "",			COLOR_DEFAULT,	COLOR_BLUE,	A_BOLD), \
+LINE(STAT_NONE,    "",			COLOR_DEFAULT,	COLOR_DEFAULT,	0), \
+LINE(STAT_STAGED,  "",			COLOR_CYAN,	COLOR_DEFAULT,	0), \
+LINE(STAT_UNSTAGED,"",			COLOR_YELLOW,	COLOR_DEFAULT,	0), \
+LINE(STAT_UNTRACKED,"",			COLOR_MAGENTA,	COLOR_DEFAULT,	0)
 
 enum line_type {
 #define LINE(type, line, fg, bg, attr) \
@@ -690,6 +719,7 @@ static struct keybinding default_keybindings[] = {
 	{ 'f',		REQ_VIEW_BLOB },
 	{ 'p',		REQ_VIEW_PAGER },
 	{ 'h',		REQ_VIEW_HELP },
+	{ 'S',		REQ_VIEW_STATUS },
 
 	/* View manipulation */
 	{ 'q',		REQ_VIEW_CLOSE },
@@ -730,9 +760,6 @@ static struct keybinding default_keybindings[] = {
 	{ 'g',		REQ_TOGGLE_REV_GRAPH },
 	{ ':',		REQ_PROMPT },
 
-	/* wgetch() with nodelay() enabled returns ERR when there's no input. */
-	{ ERR,		REQ_NONE },
-
 	/* Using the ncurses SIGWINCH handler. */
 	{ KEY_RESIZE,	REQ_SCREEN_RESIZE },
 };
@@ -745,7 +772,8 @@ static struct keybinding default_keybindings[] = {
 	KEYMAP_(TREE), \
 	KEYMAP_(BLOB), \
 	KEYMAP_(PAGER), \
-	KEYMAP_(HELP) \
+	KEYMAP_(HELP), \
+	KEYMAP_(STATUS)
 
 enum keymap {
 #define KEYMAP_(name) KEYMAP_##name
@@ -767,10 +795,9 @@ static struct keybinding *keybindings[ARRAY_SIZE(keymap_table)];
 static void
 add_keybinding(enum keymap keymap, enum request request, int key)
 {
-	struct keybinding *keybinding = keybindings[keymap];
+	struct keybinding *keybinding;
 
-	if (!keybinding)
-		keybinding = calloc(1, sizeof(*keybinding));
+	keybinding = calloc(1, sizeof(*keybinding));
 	if (!keybinding)
 		die("Failed to allocate keybinding");
 
@@ -861,7 +888,7 @@ get_key(enum request request)
 	static char buf[BUFSIZ];
 	static char key_char[] = "'X'";
 	size_t pos = 0;
-	char *sep = "    ";
+	char *sep = "";
 	int i;
 
 	buf[pos] = 0;
@@ -1222,10 +1249,12 @@ struct view {
 struct view_ops {
 	/* What type of content being displayed. Used in the title bar. */
 	const char *type;
-	/* Draw one line; @lineno must be < view->height. */
-	bool (*draw)(struct view *view, struct line *line, unsigned int lineno, bool selected);
+	/* Open and reads in all view content. */
+	bool (*open)(struct view *view);
 	/* Read one line; updates view->line. */
 	bool (*read)(struct view *view, char *data);
+	/* Draw one line; @lineno must be < view->height. */
+	bool (*draw)(struct view *view, struct line *line, unsigned int lineno, bool selected);
 	/* Depending on view, change display based on current line. */
 	bool (*enter)(struct view *view, struct line *line);
 	/* Search for regex in a line. */
@@ -1238,6 +1267,8 @@ static struct view_ops pager_ops;
 static struct view_ops main_ops;
 static struct view_ops tree_ops;
 static struct view_ops blob_ops;
+static struct view_ops help_ops;
+static struct view_ops status_ops;
 
 #define VIEW_STR(name, cmd, env, ref, ops, map) \
 	{ name, cmd, #env, ref, ops, map}
@@ -1247,13 +1278,14 @@ static struct view_ops blob_ops;
 
 
 static struct view views[] = {
-	VIEW_(MAIN,  "main",  &main_ops,  ref_head),
-	VIEW_(DIFF,  "diff",  &pager_ops, ref_commit),
-	VIEW_(LOG,   "log",   &pager_ops, ref_head),
-	VIEW_(TREE,  "tree",  &tree_ops,  ref_commit),
-	VIEW_(BLOB,  "blob",  &blob_ops,  ref_blob),
-	VIEW_(HELP,  "help",  &pager_ops, "static"),
-	VIEW_(PAGER, "pager", &pager_ops, "static"),
+	VIEW_(MAIN,   "main",   &main_ops,   ref_head),
+	VIEW_(DIFF,   "diff",   &pager_ops,  ref_commit),
+	VIEW_(LOG,    "log",    &pager_ops,  ref_head),
+	VIEW_(TREE,   "tree",   &tree_ops,   ref_commit),
+	VIEW_(BLOB,   "blob",   &blob_ops,   ref_blob),
+	VIEW_(HELP,   "help",   &help_ops,  ""),
+	VIEW_(PAGER,  "pager",  &pager_ops,  ""),
+	VIEW_(STATUS, "status", &status_ops, ""),
 };
 
 #define VIEW(req) (&views[(req) - REQ_OFFSET - 1])
@@ -1616,7 +1648,8 @@ move_view(struct view *view, enum request request)
 	}
 
 	if (!view_is_displayed(view)) {
-		view->offset += steps;
+		view->offset += scroll_steps;
+		assert(0 <= view->offset && view->offset < view->lines);
 		view->ops->select(view, &view->line[view->lineno]);
 		return;
 	}
@@ -1773,17 +1806,21 @@ begin_update(struct view *view)
 	if (opt_cmd[0]) {
 		string_copy(view->cmd, opt_cmd);
 		opt_cmd[0] = 0;
-		/* When running random commands, the view ref could have become
-		 * invalid so clear it. */
-		view->ref[0] = 0;
+		/* When running random commands, initially show the
+		 * command in the title. However, it maybe later be
+		 * overwritten if a commit line is selected. */
+		string_copy(view->ref, view->cmd);
 
 	} else if (view == VIEW(REQ_VIEW_TREE)) {
 		const char *format = view->cmd_env ? view->cmd_env : view->cmd_fmt;
+		char path[SIZEOF_STR];
 
 		if (strcmp(view->vid, view->id))
-			opt_path[0] = 0;
+			opt_path[0] = path[0] = 0;
+		else if (sq_quote(path, 0, opt_path) >= sizeof(path))
+			return FALSE;
 
-		if (!string_format(view->cmd, format, id, opt_path))
+		if (!string_format(view->cmd, format, id, path))
 			return FALSE;
 
 	} else {
@@ -1791,6 +1828,12 @@ begin_update(struct view *view)
 
 		if (!string_format(view->cmd, format, id, id, id, id, id))
 			return FALSE;
+
+		/* Put the current ref_* value to the view title ref
+		 * member. This is needed by the blob view. Most other
+		 * views sets it automatically after loading because the
+		 * first line is a commit line. */
+		string_copy(view->ref, id);
 	}
 
 	/* Special case for the pager view. */
@@ -1809,7 +1852,7 @@ begin_update(struct view *view)
 	view->offset = 0;
 	view->lines  = 0;
 	view->lineno = 0;
-	string_copy(view->vid, id);
+	string_copy_rev(view->vid, id);
 
 	if (view->line) {
 		int i;
@@ -1921,6 +1964,12 @@ update_view(struct view *view)
 		if (redraw_from > 0)
 			redraw_from--;
 
+		/* Since revision graph visualization requires knowledge
+		 * about the parent commit, it causes a further one-off
+		 * needed to be redrawn for incremental updates. */
+		if (redraw_from > 0 && opt_rev_graph)
+			redraw_from--;
+
 		/* Incrementally draw avoids flickering. */
 		redraw_view_from(view, redraw_from);
 	}
@@ -1945,53 +1994,37 @@ alloc_error:
 	report("Allocation failure");
 
 end:
+	view->ops->read(view, NULL);
 	end_update(view);
 	return FALSE;
 }
 
+static struct line *
+add_line_data(struct view *view, void *data, enum line_type type)
+{
+	struct line *line = &view->line[view->lines++];
+
+	memset(line, 0, sizeof(*line));
+	line->type = type;
+	line->data = data;
+
+	return line;
+}
+
+static struct line *
+add_line_text(struct view *view, char *data, enum line_type type)
+{
+	if (data)
+		data = strdup(data);
+
+	return data ? add_line_data(view, data, type) : NULL;
+}
+
 
 /*
  * View opening
  */
 
-static void open_help_view(struct view *view)
-{
-	char buf[BUFSIZ];
-	int lines = ARRAY_SIZE(req_info) + 2;
-	int i;
-
-	if (view->lines > 0)
-		return;
-
-	for (i = 0; i < ARRAY_SIZE(req_info); i++)
-		if (!req_info[i].request)
-			lines++;
-
-	view->line = calloc(lines, sizeof(*view->line));
-	if (!view->line) {
-		report("Allocation failure");
-		return;
-	}
-
-	view->ops->read(view, "Quick reference for tig keybindings:");
-
-	for (i = 0; i < ARRAY_SIZE(req_info); i++) {
-		char *key;
-
-		if (!req_info[i].request) {
-			view->ops->read(view, "");
-			view->ops->read(view, req_info[i].help);
-			continue;
-		}
-
-		key = get_key(req_info[i].request);
-		if (!string_format(buf, "%-25s %s", key, req_info[i].help))
-			continue;
-
-		view->ops->read(view, buf);
-	}
-}
-
 enum open_flags {
 	OPEN_DEFAULT = 0,	/* Use default view switching. */
 	OPEN_SPLIT = 1,		/* Split current view. */
@@ -2014,8 +2047,11 @@ open_view(struct view *prev, enum request request, enum open_flags flags)
 		return;
 	}
 
-	if (view == VIEW(REQ_VIEW_HELP)) {
-		open_help_view(view);
+	if (view->ops->open) {
+		if (!view->ops->open(view)) {
+			report("Failed to load %s view", view->name);
+			return;
+		}
 
 	} else if ((reload || strcmp(view->vid, view->id)) &&
 		   !begin_update(view)) {
@@ -2103,16 +2139,28 @@ view_driver(struct view *view, enum request request)
 
 	case REQ_VIEW_BLOB:
 		if (!ref_blob[0]) {
-			report("No file chosen, press 't' to open tree view");
+			report("No file chosen, press %s to open tree view",
+			       get_key(REQ_VIEW_TREE));
 			break;
 		}
-		/* Fall-through */
+		open_view(view, request, OPEN_DEFAULT);
+		break;
+
+	case REQ_VIEW_PAGER:
+		if (!VIEW(REQ_VIEW_PAGER)->lines) {
+			report("No pager content, press %s to run command from prompt",
+			       get_key(REQ_PROMPT));
+			break;
+		}
+		open_view(view, request, OPEN_DEFAULT);
+		break;
+
 	case REQ_VIEW_MAIN:
 	case REQ_VIEW_DIFF:
 	case REQ_VIEW_LOG:
 	case REQ_VIEW_TREE:
 	case REQ_VIEW_HELP:
-	case REQ_VIEW_PAGER:
+	case REQ_VIEW_STATUS:
 		open_view(view, request, OPEN_DEFAULT);
 		break;
 
@@ -2192,7 +2240,7 @@ view_driver(struct view *view, enum request request)
 		break;
 
 	case REQ_SHOW_VERSION:
-		report("%s (built %s)", VERSION, __DATE__);
+		report("tig-%s (built %s)", VERSION, __DATE__);
 		return TRUE;
 
 	case REQ_SCREEN_RESIZE:
@@ -2315,7 +2363,7 @@ add_describe_ref(char *buf, size_t *bufpos, char *commit_id, const char *sep)
 	char *ref = NULL;
 	FILE *pipe;
 
-	if (!string_format(refbuf, "git describe %s", commit_id))
+	if (!string_format(refbuf, "git describe %s 2>/dev/null", commit_id))
 		return TRUE;
 
 	pipe = popen(refbuf, "r");
@@ -2357,7 +2405,8 @@ add_pager_refs(struct view *view, struct line *line)
 
 	do {
 		struct ref *ref = refs[refpos];
-		char *fmt = ref->tag ? "%s[%s]" : "%s%s";
+		char *fmt = ref->tag    ? "%s[%s]" :
+		            ref->remote ? "%s<%s>" : "%s%s";
 
 		if (!string_format_from(buf, &bufpos, fmt, sep, ref->name))
 			return;
@@ -2379,27 +2428,21 @@ try_add_describe_ref:
 	if (!realloc_lines(view, view->line_size + 1))
 		return;
 
-	line = &view->line[view->lines];
-	line->data = strdup(buf);
-	if (!line->data)
-		return;
-
-	line->type = LINE_PP_REFS;
-	view->lines++;
+	add_line_text(view, buf, LINE_PP_REFS);
 }
 
 static bool
 pager_read(struct view *view, char *data)
 {
-	struct line *line = &view->line[view->lines];
+	struct line *line;
 
-	line->data = strdup(data);
-	if (!line->data)
+	if (!data)
+		return TRUE;
+
+	line = add_line_text(view, data, get_line_type(data));
+	if (!line)
 		return FALSE;
 
-	line->type = get_line_type(line->data);
-	view->lines++;
-
 	if (line->type == LINE_COMMIT &&
 	    (view == VIEW(REQ_VIEW_DIFF) ||
 	     view == VIEW(REQ_VIEW_LOG)))
@@ -2453,17 +2496,73 @@ static void
 pager_select(struct view *view, struct line *line)
 {
 	if (line->type == LINE_COMMIT) {
-		char *text = line->data;
+		char *text = line->data + STRING_SIZE("commit ");
 
-		string_copy(view->ref, text + STRING_SIZE("commit "));
-		string_copy(ref_commit, view->ref);
+		if (view != VIEW(REQ_VIEW_PAGER))
+			string_copy_rev(view->ref, text);
+		string_copy_rev(ref_commit, text);
 	}
 }
 
 static struct view_ops pager_ops = {
 	"line",
-	pager_draw,
+	NULL,
 	pager_read,
+	pager_draw,
+	pager_enter,
+	pager_grep,
+	pager_select,
+};
+
+
+/*
+ * Help backend
+ */
+
+static bool
+help_open(struct view *view)
+{
+	char buf[BUFSIZ];
+	int lines = ARRAY_SIZE(req_info) + 2;
+	int i;
+
+	if (view->lines > 0)
+		return TRUE;
+
+	for (i = 0; i < ARRAY_SIZE(req_info); i++)
+		if (!req_info[i].request)
+			lines++;
+
+	view->line = calloc(lines, sizeof(*view->line));
+	if (!view->line)
+		return FALSE;
+
+	add_line_text(view, "Quick reference for tig keybindings:", LINE_DEFAULT);
+
+	for (i = 0; i < ARRAY_SIZE(req_info); i++) {
+		char *key;
+
+		if (!req_info[i].request) {
+			add_line_text(view, "", LINE_DEFAULT);
+			add_line_text(view, req_info[i].help, LINE_DEFAULT);
+			continue;
+		}
+
+		key = get_key(req_info[i].request);
+		if (!string_format(buf, "    %-25s %s", key, req_info[i].help))
+			continue;
+
+		add_line_text(view, buf, LINE_DEFAULT);
+	}
+
+	return TRUE;
+}
+
+static struct view_ops help_ops = {
+	"line",
+	help_open,
+	NULL,
+	pager_draw,
 	pager_enter,
 	pager_grep,
 	pager_select,
@@ -2503,7 +2602,7 @@ tree_compare_entry(enum line_type type1, char *name1,
 static bool
 tree_read(struct view *view, char *text)
 {
-	size_t textlen = strlen(text);
+	size_t textlen = text ? strlen(text) : 0;
 	char buf[SIZEOF_STR];
 	unsigned long pos;
 	enum line_type type;
@@ -2517,21 +2616,18 @@ tree_read(struct view *view, char *text)
 
 	if (first_read) {
 		/* Add path info line */
-		if (string_format(buf, "Directory path /%s", opt_path) &&
-		    realloc_lines(view, view->line_size + 1) &&
-		    pager_read(view, buf))
-			view->line[view->lines - 1].type = LINE_DEFAULT;
-		else
+		if (!string_format(buf, "Directory path /%s", opt_path) ||
+		    !realloc_lines(view, view->line_size + 1) ||
+		    !add_line_text(view, buf, LINE_DEFAULT))
 			return FALSE;
 
 		/* Insert "link" to parent directory. */
-		if (*opt_path &&
-		    string_format(buf, TREE_UP_FORMAT, view->ref) &&
-		    realloc_lines(view, view->line_size + 1) &&
-		    pager_read(view, buf))
-			view->line[view->lines - 1].type = LINE_TREE_DIR;
-		else if (*opt_path)
-			return FALSE;
+		if (*opt_path) {
+			if (!string_format(buf, TREE_UP_FORMAT, view->ref) ||
+			    !realloc_lines(view, view->line_size + 1) ||
+			    !add_line_text(view, buf, LINE_TREE_DIR))
+				return FALSE;
+		}
 	}
 
 	/* Strip the path part ... */
@@ -2570,21 +2666,20 @@ tree_read(struct view *view, char *text)
 		return TRUE;
 	}
 
-	if (!pager_read(view, text))
+	if (!add_line_text(view, text, type))
 		return FALSE;
 
 	/* Move the current line to the first tree entry. */
 	if (first_read)
 		view->lineno++;
 
-	view->line[view->lines - 1].type = type;
 	return TRUE;
 }
 
 static bool
 tree_enter(struct view *view, struct line *line)
 {
-	enum open_flags flags = display[0] == view ? OPEN_SPLIT : OPEN_DEFAULT;
+	enum open_flags flags;
 	enum request request;
 
 	switch (line->type) {
@@ -2614,11 +2709,12 @@ tree_enter(struct view *view, struct line *line)
 
 		/* Trees and subtrees share the same ID, so they are not not
 		 * unique like blobs. */
-		flags |= OPEN_RELOAD;
+		flags = OPEN_RELOAD;
 		request = REQ_VIEW_TREE;
 		break;
 
 	case LINE_TREE_FILE:
+		flags = display[0] == view ? OPEN_SPLIT : OPEN_DEFAULT;
 		request = REQ_VIEW_BLOB;
 		break;
 
@@ -2634,27 +2730,23 @@ tree_enter(struct view *view, struct line *line)
 static void
 tree_select(struct view *view, struct line *line)
 {
-	char *text = line->data;
-
-	text += STRING_SIZE("100644 blob ");
+	char *text = line->data + STRING_SIZE("100644 blob ");
 
 	if (line->type == LINE_TREE_FILE) {
-		string_ncopy(ref_blob, text, 40);
-		/* Also update the blob view's ref, since all there must always
-		 * be in sync. */
-		string_copy(VIEW(REQ_VIEW_BLOB)->ref, ref_blob);
+		string_copy_rev(ref_blob, text);
 
 	} else if (line->type != LINE_TREE_DIR) {
 		return;
 	}
 
-	string_ncopy(view->ref, text, 40);
+	string_copy_rev(view->ref, text);
 }
 
 static struct view_ops tree_ops = {
 	"file",
-	pager_draw,
+	NULL,
 	tree_read,
+	pager_draw,
 	tree_enter,
 	pager_grep,
 	tree_select,
@@ -2663,18 +2755,14 @@ static struct view_ops tree_ops = {
 static bool
 blob_read(struct view *view, char *line)
 {
-	bool state = pager_read(view, line);
-
-	if (state == TRUE)
-		view->line[view->lines - 1].type = LINE_DEFAULT;
-
-	return state;
+	return add_line_text(view, line, LINE_DEFAULT);
 }
 
 static struct view_ops blob_ops = {
 	"line",
-	pager_draw,
+	NULL,
 	blob_read,
+	pager_draw,
 	pager_enter,
 	pager_grep,
 	pager_select,
@@ -2682,12 +2770,350 @@ static struct view_ops blob_ops = {
 
 
 /*
- * Main view backend
+ * Status backend
+ */
+
+struct status {
+	char status;
+	struct {
+		mode_t mode;
+		char rev[SIZEOF_REV];
+	} old;
+	struct {
+		mode_t mode;
+		char rev[SIZEOF_REV];
+	} new;
+	char name[SIZEOF_STR];
+};
+
+/* Get fields from the diff line:
+ * :100644 100644 06a5d6ae9eca55be2e0e585a152e6b1336f2b20e 0000000000000000000000000000000000000000 M
+ */
+static inline bool
+status_get_diff(struct status *file, char *buf, size_t bufsize)
+{
+	char *old_mode = buf +  1;
+	char *new_mode = buf +  8;
+	char *old_rev  = buf + 15;
+	char *new_rev  = buf + 56;
+	char *status   = buf + 97;
+
+	if (bufsize != 99 ||
+	    old_mode[-1] != ':' ||
+	    new_mode[-1] != ' ' ||
+	    old_rev[-1]  != ' ' ||
+	    new_rev[-1]  != ' ' ||
+	    status[-1]   != ' ')
+		return FALSE;
+
+	file->status = *status;
+
+	string_copy_rev(file->old.rev, old_rev);
+	string_copy_rev(file->new.rev, new_rev);
+
+	file->old.mode = strtoul(old_mode, NULL, 8);
+	file->new.mode = strtoul(new_mode, NULL, 8);
+
+	file->name[0] = 0;
+
+	return TRUE;
+}
+
+static bool
+status_run(struct view *view, const char cmd[], bool diff, enum line_type type)
+{
+	struct status *file = NULL;
+	char buf[SIZEOF_STR * 4];
+	size_t bufsize = 0;
+	FILE *pipe;
+
+	pipe = popen(cmd, "r");
+	if (!pipe)
+		return FALSE;
+
+	add_line_data(view, NULL, type);
+
+	while (!feof(pipe) && !ferror(pipe)) {
+		char *sep;
+		size_t readsize;
+
+		readsize = fread(buf + bufsize, 1, sizeof(buf) - bufsize, pipe);
+		if (!readsize)
+			break;
+		bufsize += readsize;
+
+		/* Process while we have NUL chars. */
+		while ((sep = memchr(buf, 0, bufsize))) {
+			size_t sepsize = sep - buf + 1;
+
+			if (!file) {
+				if (!realloc_lines(view, view->line_size + 1))
+					goto error_out;
+
+				file = calloc(1, sizeof(*file));
+				if (!file)
+					goto error_out;
+
+				add_line_data(view, file, type);
+			}
+
+			/* Parse diff info part. */
+			if (!diff) {
+				file->status = '?';
+
+			} else if (!file->status) {
+				if (!status_get_diff(file, buf, sepsize))
+					goto error_out;
+
+				bufsize -= sepsize;
+				memmove(buf, sep + 1, bufsize);
+
+				sep = memchr(buf, 0, bufsize);
+				if (!sep)
+					break;
+				sepsize = sep - buf + 1;
+			}
+
+			/* git-ls-files just delivers a NUL separated
+			 * list of file names similar to the second half
+			 * of the git-diff-* output. */
+			string_ncopy(file->name, buf, sepsize);
+			bufsize -= sepsize;
+			memmove(buf, sep + 1, bufsize);
+			file = NULL;
+		}
+	}
+
+	if (ferror(pipe)) {
+error_out:
+		pclose(pipe);
+		return FALSE;
+	}
+
+	if (!view->line[view->lines - 1].data)
+		add_line_data(view, NULL, LINE_STAT_NONE);
+
+	pclose(pipe);
+	return TRUE;
+}
+
+#define STATUS_DIFF_INDEX_CMD "git diff-index -z --cached HEAD"
+#define STATUS_DIFF_FILES_CMD "git diff-files -z"
+#define STATUS_LIST_OTHER_CMD \
+	"_git_exclude=$(git rev-parse --git-dir)/info/exclude;" \
+	"test -f \"$_git_exclude\" && exclude=\"--exclude-from=$_git_exclude\";" \
+	"git ls-files -z --others --exclude-per-directory=.gitignore \"$exclude\"" \
+
+/* First parse staged info using git-diff-index(1), then parse unstaged
+ * info using git-diff-files(1), and finally untracked files using
+ * git-ls-files(1). */
+static bool
+status_open(struct view *view)
+{
+	size_t i;
+
+	for (i = 0; i < view->lines; i++)
+		free(view->line[i].data);
+	free(view->line);
+	view->lines = view->line_size = 0;
+	view->line = NULL;
+
+	if (!realloc_lines(view, view->line_size + 6))
+		return FALSE;
+
+	if (!status_run(view, STATUS_DIFF_INDEX_CMD, TRUE, LINE_STAT_STAGED) ||
+	    !status_run(view, STATUS_DIFF_FILES_CMD, TRUE, LINE_STAT_UNSTAGED) ||
+	    !status_run(view, STATUS_LIST_OTHER_CMD, FALSE, LINE_STAT_UNTRACKED))
+		return FALSE;
+
+	return TRUE;
+}
+
+static bool
+status_draw(struct view *view, struct line *line, unsigned int lineno, bool selected)
+{
+	struct status *status = line->data;
+
+	wmove(view->win, lineno, 0);
+
+	if (selected) {
+		wattrset(view->win, get_line_attr(LINE_CURSOR));
+		wchgat(view->win, -1, 0, LINE_CURSOR, NULL);
+
+	} else if (!status && line->type != LINE_STAT_NONE) {
+		wattrset(view->win, get_line_attr(LINE_STAT_SECTION));
+		wchgat(view->win, -1, 0, LINE_STAT_SECTION, NULL);
+
+	} else {
+		wattrset(view->win, get_line_attr(line->type));
+	}
+
+	if (!status) {
+		char *text;
+
+		switch (line->type) {
+		case LINE_STAT_STAGED:
+			text = "Changes to be committed:";
+			break;
+
+		case LINE_STAT_UNSTAGED:
+			text = "Changed but not updated:";
+			break;
+
+		case LINE_STAT_UNTRACKED:
+			text = "Untracked files:";
+			break;
+
+		case LINE_STAT_NONE:
+			text = "    (no files)";
+			break;
+
+		default:
+			return FALSE;
+		}
+
+		waddstr(view->win, text);
+		return TRUE;
+	}
+
+	waddch(view->win, status->status);
+	if (!selected)
+		wattrset(view->win, A_NORMAL);
+	wmove(view->win, lineno, 4);
+	waddstr(view->win, status->name);
+
+	return TRUE;
+}
+
+static bool
+status_enter(struct view *view, struct line *line)
+{
+	struct status *status = line->data;
+	const char *cmd;
+	char buf[SIZEOF_STR];
+	size_t bufsize = 0;
+	size_t written = 0;
+	FILE *pipe;
+
+	if (!status)
+		return TRUE;
+
+	switch (line->type) {
+	case LINE_STAT_STAGED:
+		if (!string_format_from(buf, &bufsize, "%06o %s\t%s%c",
+				        status->old.mode,
+					status->old.rev,
+					status->name, 0))
+			return FALSE;
+		cmd = "git update-index -z --index-info";
+		break;
+
+	case LINE_STAT_UNSTAGED:
+	case LINE_STAT_UNTRACKED:
+		if (!string_format_from(buf, &bufsize, "%s%c", status->name, 0))
+			return FALSE;
+		cmd = "git update-index -z --add --remove --stdin";
+		break;
+
+	default:
+		die("w00t");
+	}
+
+	pipe = popen(cmd, "w");
+	if (!pipe)
+		return FALSE;
+
+	while (!ferror(pipe) && written < bufsize) {
+		written += fwrite(buf + written, 1, bufsize - written, pipe);
+	}
+
+	pclose(pipe);
+
+	if (written != bufsize)
+		return FALSE;
+
+	open_view(view, REQ_VIEW_STATUS, OPEN_RELOAD);
+	return TRUE;
+}
+
+static void
+status_select(struct view *view, struct line *line)
+{
+	char *text;
+
+	switch (line->type) {
+	case LINE_STAT_STAGED:
+		text = "Press Enter to unstage file for commit";
+		break;
+
+	case LINE_STAT_UNSTAGED:
+		text = "Press Enter to stage file for commit  ";
+		break;
+
+	case LINE_STAT_UNTRACKED:
+		text = "Press Enter to stage file for addition";
+		break;
+
+	case LINE_STAT_NONE:
+		return;
+
+	default:
+		die("w00t");
+	}
+
+	string_ncopy(view->ref, text, strlen(text));
+}
+
+static bool
+status_grep(struct view *view, struct line *line)
+{
+	struct status *status = line->data;
+	enum { S_STATUS, S_NAME, S_END } state;
+	char buf[2] = "?";
+	regmatch_t pmatch;
+
+	if (!status)
+		return FALSE;
+
+	for (state = S_STATUS; state < S_END; state++) {
+		char *text;
+
+		switch (state) {
+		case S_NAME:	text = status->name;	break;
+		case S_STATUS:
+			buf[0] = status->status;
+			text = buf;
+			break;
+
+		default:
+			return FALSE;
+		}
+
+		if (regexec(view->regex, text, 1, &pmatch, 0) != REG_NOMATCH)
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
+static struct view_ops status_ops = {
+	"file",
+	status_open,
+	NULL,
+	status_draw,
+	status_enter,
+	status_grep,
+	status_select,
+};
+
+
+/*
+ * Revision graph
  */
 
 struct commit {
 	char id[SIZEOF_REV];		/* SHA1 ID. */
-	char title[75];			/* First line of the commit message. */
+	char title[128];		/* First line of the commit message. */
 	char author[75];		/* Author of the commit. */
 	struct tm time;			/* Date from the author ident. */
 	struct ref **refs;		/* Repository references. */
@@ -2695,6 +3121,198 @@ struct commit {
 	size_t graph_size;		/* The width of the graph array. */
 };
 
+/* Size of rev graph with no  "padding" columns */
+#define SIZEOF_REVITEMS	(SIZEOF_REVGRAPH - (SIZEOF_REVGRAPH / 2))
+
+struct rev_graph {
+	struct rev_graph *prev, *next, *parents;
+	char rev[SIZEOF_REVITEMS][SIZEOF_REV];
+	size_t size;
+	struct commit *commit;
+	size_t pos;
+};
+
+/* Parents of the commit being visualized. */
+static struct rev_graph graph_parents[4];
+
+/* The current stack of revisions on the graph. */
+static struct rev_graph graph_stacks[4] = {
+	{ &graph_stacks[3], &graph_stacks[1], &graph_parents[0] },
+	{ &graph_stacks[0], &graph_stacks[2], &graph_parents[1] },
+	{ &graph_stacks[1], &graph_stacks[3], &graph_parents[2] },
+	{ &graph_stacks[2], &graph_stacks[0], &graph_parents[3] },
+};
+
+static inline bool
+graph_parent_is_merge(struct rev_graph *graph)
+{
+	return graph->parents->size > 1;
+}
+
+static inline void
+append_to_rev_graph(struct rev_graph *graph, chtype symbol)
+{
+	struct commit *commit = graph->commit;
+
+	if (commit->graph_size < ARRAY_SIZE(commit->graph) - 1)
+		commit->graph[commit->graph_size++] = symbol;
+}
+
+static void
+done_rev_graph(struct rev_graph *graph)
+{
+	if (graph_parent_is_merge(graph) &&
+	    graph->pos < graph->size - 1 &&
+	    graph->next->size == graph->size + graph->parents->size - 1) {
+		size_t i = graph->pos + graph->parents->size - 1;
+
+		graph->commit->graph_size = i * 2;
+		while (i < graph->next->size - 1) {
+			append_to_rev_graph(graph, ' ');
+			append_to_rev_graph(graph, '\\');
+			i++;
+		}
+	}
+
+	graph->size = graph->pos = 0;
+	graph->commit = NULL;
+	memset(graph->parents, 0, sizeof(*graph->parents));
+}
+
+static void
+push_rev_graph(struct rev_graph *graph, char *parent)
+{
+	int i;
+
+	/* "Collapse" duplicate parents lines.
+	 *
+	 * FIXME: This needs to also update update the drawn graph but
+	 * for now it just serves as a method for pruning graph lines. */
+	for (i = 0; i < graph->size; i++)
+		if (!strncmp(graph->rev[i], parent, SIZEOF_REV))
+			return;
+
+	if (graph->size < SIZEOF_REVITEMS) {
+		string_ncopy(graph->rev[graph->size++], parent, SIZEOF_REV);
+	}
+}
+
+static chtype
+get_rev_graph_symbol(struct rev_graph *graph)
+{
+	chtype symbol;
+
+	if (graph->parents->size == 0)
+		symbol = REVGRAPH_INIT;
+	else if (graph_parent_is_merge(graph))
+		symbol = REVGRAPH_MERGE;
+	else if (graph->pos >= graph->size)
+		symbol = REVGRAPH_BRANCH;
+	else
+		symbol = REVGRAPH_COMMIT;
+
+	return symbol;
+}
+
+static void
+draw_rev_graph(struct rev_graph *graph)
+{
+	struct rev_filler {
+		chtype separator, line;
+	};
+	enum { DEFAULT, RSHARP, RDIAG, LDIAG };
+	static struct rev_filler fillers[] = {
+		{ ' ',	REVGRAPH_LINE },
+		{ '`',	'.' },
+		{ '\'',	' ' },
+		{ '/',	' ' },
+	};
+	chtype symbol = get_rev_graph_symbol(graph);
+	struct rev_filler *filler;
+	size_t i;
+
+	filler = &fillers[DEFAULT];
+
+	for (i = 0; i < graph->pos; i++) {
+		append_to_rev_graph(graph, filler->line);
+		if (graph_parent_is_merge(graph->prev) &&
+		    graph->prev->pos == i)
+			filler = &fillers[RSHARP];
+
+		append_to_rev_graph(graph, filler->separator);
+	}
+
+	/* Place the symbol for this revision. */
+	append_to_rev_graph(graph, symbol);
+
+	if (graph->prev->size > graph->size)
+		filler = &fillers[RDIAG];
+	else
+		filler = &fillers[DEFAULT];
+
+	i++;
+
+	for (; i < graph->size; i++) {
+		append_to_rev_graph(graph, filler->separator);
+		append_to_rev_graph(graph, filler->line);
+		if (graph_parent_is_merge(graph->prev) &&
+		    i < graph->prev->pos + graph->parents->size)
+			filler = &fillers[RSHARP];
+		if (graph->prev->size > graph->size)
+			filler = &fillers[LDIAG];
+	}
+
+	if (graph->prev->size > graph->size) {
+		append_to_rev_graph(graph, filler->separator);
+		if (filler->line != ' ')
+			append_to_rev_graph(graph, filler->line);
+	}
+}
+
+/* Prepare the next rev graph */
+static void
+prepare_rev_graph(struct rev_graph *graph)
+{
+	size_t i;
+
+	/* First, traverse all lines of revisions up to the active one. */
+	for (graph->pos = 0; graph->pos < graph->size; graph->pos++) {
+		if (!strcmp(graph->rev[graph->pos], graph->commit->id))
+			break;
+
+		push_rev_graph(graph->next, graph->rev[graph->pos]);
+	}
+
+	/* Interleave the new revision parent(s). */
+	for (i = 0; i < graph->parents->size; i++)
+		push_rev_graph(graph->next, graph->parents->rev[i]);
+
+	/* Lastly, put any remaining revisions. */
+	for (i = graph->pos + 1; i < graph->size; i++)
+		push_rev_graph(graph->next, graph->rev[i]);
+}
+
+static void
+update_rev_graph(struct rev_graph *graph)
+{
+	/* If this is the finalizing update ... */
+	if (graph->commit)
+		prepare_rev_graph(graph);
+
+	/* Graph visualization needs a one rev look-ahead,
+	 * so the first update doesn't visualize anything. */
+	if (!graph->prev->commit)
+		return;
+
+	draw_rev_graph(graph->prev);
+	done_rev_graph(graph->prev->prev);
+}
+
+
+/*
+ * Main view backend
+ */
+
 static bool
 main_draw(struct view *view, struct line *line, unsigned int lineno, bool selected)
 {
@@ -2762,6 +3380,7 @@ main_draw(struct view *view, struct line *line, unsigned int lineno, bool select
 		for (i = 0; i < commit->graph_size; i++)
 			waddch(view->win, commit->graph[i]);
 
+		waddch(view->win, ' ');
 		col += commit->graph_size + 1;
 	}
 
@@ -2775,6 +3394,8 @@ main_draw(struct view *view, struct line *line, unsigned int lineno, bool select
 				;
 			else if (commit->refs[i]->tag)
 				wattrset(view->win, get_line_attr(LINE_MAIN_TAG));
+			else if (commit->refs[i]->remote)
+				wattrset(view->win, get_line_attr(LINE_MAIN_REMOTE));
 			else
 				wattrset(view->win, get_line_attr(LINE_MAIN_REF));
 			waddstr(view->win, "[");
@@ -2806,63 +3427,68 @@ main_draw(struct view *view, struct line *line, unsigned int lineno, bool select
 static bool
 main_read(struct view *view, char *line)
 {
-	enum line_type type = get_line_type(line);
-	struct commit *commit = view->lines
-			      ? view->line[view->lines - 1].data : NULL;
+	static struct rev_graph *graph = graph_stacks;
+	enum line_type type;
+	struct commit *commit;
 
-	switch (type) {
-	case LINE_COMMIT:
+	if (!line) {
+		update_rev_graph(graph);
+		return TRUE;
+	}
+
+	type = get_line_type(line);
+	if (type == LINE_COMMIT) {
 		commit = calloc(1, sizeof(struct commit));
 		if (!commit)
 			return FALSE;
 
-		line += STRING_SIZE("commit ");
-
-		view->line[view->lines++].data = commit;
-		string_copy(commit->id, line);
+		string_copy_rev(commit->id, line + STRING_SIZE("commit "));
 		commit->refs = get_refs(commit->id);
-		commit->graph[commit->graph_size++] = ACS_LTEE;
+		graph->commit = commit;
+		add_line_data(view, commit, LINE_MAIN_COMMIT);
+		return TRUE;
+	}
+
+	if (!view->lines)
+		return TRUE;
+	commit = view->line[view->lines - 1].data;
+
+	switch (type) {
+	case LINE_PARENT:
+		push_rev_graph(graph->parents, line + STRING_SIZE("parent "));
 		break;
 
 	case LINE_AUTHOR:
 	{
+		/* Parse author lines where the name may be empty:
+		 *	author  <email@address.tld> 1138474660 +0100
+		 */
 		char *ident = line + STRING_SIZE("author ");
-		char *end = strchr(ident, '<');
+		char *nameend = strchr(ident, '<');
+		char *emailend = strchr(ident, '>');
 
-		if (!commit)
+		if (!nameend || !emailend)
 			break;
 
-		if (end) {
-			char *email = end + 1;
+		update_rev_graph(graph);
+		graph = graph->next;
 
-			for (; end > ident && isspace(end[-1]); end--) ;
-
-			if (end == ident && *email) {
-				ident = email;
-				end = strchr(ident, '>');
-				for (; end > ident && isspace(end[-1]); end--) ;
-			}
-			*end = 0;
+		*nameend = *emailend = 0;
+		ident = chomp_string(ident);
+		if (!*ident) {
+			ident = chomp_string(nameend + 1);
+			if (!*ident)
+				ident = "Unknown";
 		}
 
-		/* End is NULL or ident meaning there's no author. */
-		if (end <= ident)
-			ident = "Unknown";
-
 		string_copy(commit->author, ident);
 
 		/* Parse epoch and timezone */
-		if (end) {
-			char *secs = strchr(end + 1, '>');
-			char *zone;
-			time_t time;
+		if (emailend[1] == ' ') {
+			char *secs = emailend + 2;
+			char *zone = strchr(secs, ' ');
+			time_t time = (time_t) atol(secs);
 
-			if (!secs || secs[1] != ' ')
-				break;
-
-			secs += 2;
-			time = (time_t) atol(secs);
-			zone = strchr(secs, ' ');
 			if (zone && strlen(zone) == STRING_SIZE(" +0700")) {
 				long tz;
 
@@ -2877,27 +3503,31 @@ main_read(struct view *view, char *line)
 
 				time -= tz;
 			}
+
 			gmtime_r(&time, &commit->time);
 		}
 		break;
 	}
 	default:
-		if (!commit)
-			break;
-
 		/* Fill in the commit title if it has not already been set. */
 		if (commit->title[0])
 			break;
 
 		/* Require titles to start with a non-space character at the
 		 * offset used by git log. */
-		/* FIXME: More gracefull handling of titles; append "..." to
-		 * shortened titles, etc. */
-		if (strncmp(line, "    ", 4) ||
-		    isspace(line[4]))
+		if (strncmp(line, "    ", 4))
 			break;
+		line += 4;
+		/* Well, if the title starts with a whitespace character,
+		 * try to be forgiving.  Otherwise we end up with no title. */
+		while (isspace(*line))
+			line++;
+		if (*line == '\0')
+			break;
+		/* FIXME: More graceful handling of titles; append "..." to
+		 * shortened titles, etc. */
 
-		string_copy(commit->title, line + 4);
+		string_copy(commit->title, line);
 	}
 
 	return TRUE;
@@ -2948,14 +3578,15 @@ main_select(struct view *view, struct line *line)
 {
 	struct commit *commit = line->data;
 
-	string_copy(view->ref, commit->id);
-	string_copy(ref_commit, view->ref);
+	string_copy_rev(view->ref, commit->id);
+	string_copy_rev(ref_commit, view->ref);
 }
 
 static struct view_ops main_ops = {
 	"commit",
-	main_draw,
+	NULL,
 	main_read,
+	main_draw,
 	main_enter,
 	main_grep,
 	main_select,
@@ -3346,6 +3977,7 @@ read_ref(char *id, int idlen, char *name, int namelen)
 {
 	struct ref *ref;
 	bool tag = FALSE;
+	bool remote = FALSE;
 
 	if (!strncmp(name, "refs/tags/", STRING_SIZE("refs/tags/"))) {
 		/* Commits referenced by tags has "^{}" appended. */
@@ -3359,6 +3991,11 @@ read_ref(char *id, int idlen, char *name, int namelen)
 		namelen -= STRING_SIZE("refs/tags/");
 		name	+= STRING_SIZE("refs/tags/");
 
+	} else if (!strncmp(name, "refs/remotes/", STRING_SIZE("refs/remotes/"))) {
+		remote = TRUE;
+		namelen -= STRING_SIZE("refs/remotes/");
+		name	+= STRING_SIZE("refs/remotes/");
+
 	} else if (!strncmp(name, "refs/heads/", STRING_SIZE("refs/heads/"))) {
 		namelen -= STRING_SIZE("refs/heads/");
 		name	+= STRING_SIZE("refs/heads/");
@@ -3379,7 +4016,8 @@ read_ref(char *id, int idlen, char *name, int namelen)
 	strncpy(ref->name, name, namelen);
 	ref->name[namelen] = 0;
 	ref->tag = tag;
-	string_copy(ref->id, id);
+	ref->remote = remote;
+	string_copy_rev(ref->id, id);
 
 	return OK;
 }
@@ -3533,6 +4171,13 @@ main(int argc, char *argv[])
 		/* Refresh, accept single keystroke of input */
 		key = wgetch(status_win);
 
+		/* wgetch() with nodelay() enabled returns ERR when there's no
+		 * input. */
+		if (key == ERR) {
+			request = REQ_NONE;
+			continue;
+		}
+
 		request = get_keybinding(display[current_view]->keymap, key);
 
 		/* Some low-level request handling. This keeps access to
